import{T as l,a as r}from"./Tags-BCrwqmQ2.js";import{s as o,u as i,o as s,R as e,P as u}from"./index-DH3zn7Tz.js";const d=o.div`
  width: 100%;
  height: 100%;
  overflow-x: auto;
  overflow-y: hidden;
  margin: auto;
  text-align: center;
  wordbreak: normal;
  display: flex;
  flex-direction: row;
  align-content: space-around;
  justify-content: left;
  align-items: center;
  -ms-overflow-style: none; /* IE and Edge */
  scrollbar-width: none; /* Firefox */
  &::-webkit-scrollbar {
    display: none;
  }
`,m=n=>{const a=i(s({artifactAlias:n.input}));return a.loading?e.createElement(u,null):a.result==null?e.createElement("div",null,"-"):e.createElement(d,null,a.result.map(t=>e.createElement(l,{key:t,tag:{name:t,colorIndex:r.ALIAS},noun:"alias"})))};export{m as default};
//# sourceMappingURL=Component-DUleNz_E.js.map
