class Error(Exception): ...


class InternalError(Error): ...


class OpCallError(Error): ...


class WeaveTypeError(Error): ...


class WeaveSerializeError(Error): ...


class WeaveOpSerializeError(WeaveSerializeError): ...


class WeaveInitError(Error): ...


class WeaveDefinitionError(Error): ...


class WeaveWandbAuthenticationException(Error): ...
